package org.drools.linear;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class ForwardChainingEngine {
    private Map objectTypes;
    
    private Rule[] rules;

    public ForwardChainingEngine(Rule[] rules) {
        this.rules = rules;
        this.objectTypes = new HashMap();
    }

    public void assertObject( Object object ) {
        List facts = ( List ) this.objectTypes.get( object.getClass() );
        if ( facts == null ) {
            facts = new ArrayList();
            this.objectTypes.put( object.getClass(), facts );
        }
        
        facts.add( object );
        
        // We only want to attempt evaluations on rules that depend on this ObjectType
        Rule[] rules = getDependantRules( object.getClass() );        
        for ( int i = 0; i < rules.length; i++ ) {
            Class[] classes = rules[i].getObjectTypes();
            Object[] objects = new Object[ classes.length ];
            
            // A class can be in more than one column, so we return those positions as an int[]
            int[] indexes = getClassIndexes( classes, object.getClass() );
            
            // So we make sure for each column we attempt a cross product
            // It is this feature that allows a column to match against a same type column
            for ( int j = 0; j < indexes.length; j++ ) {
                doCrossProduct( rules[i], object, indexes[j], classes, objects, 0);   
            }                        
        }
    }    
    
    /**
     * Returns a Rule[] for Rules that depend on the given Class
     * @param clazz
     * @return
     */
    public Rule[] getDependantRules(Class clazz) {
        List list = new ArrayList();
        
        for ( int i = 0; i < this.rules.length; i++ ) {
            Class[] classes = this.rules[i].getObjectTypes();
            for ( int j = 0; j < classes.length; j++ ) {
                if ( classes[j] == clazz ) {
                    list.add( this.rules[i] );
                    break;
                }
            }
        }        
        return ( Rule[] ) list.toArray( new Rule[ list.size() ] );
    }
    
    public void doCrossProduct(Rule rule, Object object, int objectIndex, Class[] classes, Object[] objects, int currentIndex) {
        // We don't have have enough assert object types to satisfy this rule, so return.
        if ( !isSatisfied( rule ) ) {
            return;
        }
        
        // If the currentIndex is the same as the objectIndex we want to
        // skip all the facts in memory, instead using the passed object
        if ( objectIndex == currentIndex ) {
            Object[] newObjects = new Object[objects.length];
            System.arraycopy( objects, 0, newObjects, 0, currentIndex );
            newObjects[currentIndex] = object;
            
            // Keep building up the cross product object[] 
            // When its full then evaluate it.            
            if ( currentIndex + 1 != classes.length ) {            
                doCrossProduct( rule, object, objectIndex, classes, newObjects, currentIndex+1);
            } else {
                rule.evaluate( newObjects );
            }                        
        } else {
        
            List list = (List) this.objectTypes.get( classes[currentIndex] );
            for ( Iterator it = list.iterator(); it.hasNext(); ) {            
                Object[] newObjects = new Object[objects.length];
                System.arraycopy( objects, 0, newObjects, 0, currentIndex );
                
                Object newObject = it.next();
                // I we are joining with ourselves abort this Object[]
                if ( object == newObject ) {
                    return;
                }
                newObjects[currentIndex] = newObject;
                
                // Keep building up the cross product object[] 
                // When its full then evaluate it.
                if ( currentIndex + 1 != classes.length ) {            
                    doCrossProduct( rule, object, objectIndex, classes, newObjects, currentIndex+1);
                } else {
                    rule.evaluate( newObjects );
                }
            }    
        }
    }   
    
    /**
     * Returns the index positions of the Class in the Class[]
     * @param classes
     * @param clazz
     * @return
     */
    public int[] getClassIndexes(Class[] classes, Class clazz) {
        // Build up a list of the index positions of t his Class
        List list = new ArrayList();
        for ( int i = 0; i < classes.length; i++ ){
            if ( classes[i] == clazz ) {
                list.add( new Integer( i ) );
            }
        }
        
        // Convert the List to an int[]
        int[] indexes = new int[ list.size() ];
        int i = 0;
        for ( Iterator it = list.iterator(); it.hasNext(); ) {
            indexes[i++] = ( ( Integer ) it.next() ).intValue();
        }
        
        return indexes;
    }
    
    /**
     * Checks if the Rule can be satisfied by the current assertect objects and their ObjectTypes
     * @param rule
     * @return
     */
    public boolean isSatisfied(Rule rule) {
        Class[] classes = rule.getObjectTypes();
        for ( int i = 0; i < classes.length; i++ ) {
            if ( this.objectTypes.get( classes[i] ) == null ) {
                return false;
            }
        }
        return true;
    }    
}
